C
C  $$ NEWKMEANS.FOR
C
C **************************************************************************
C *  SPIDER - MODULAR IMAGE PROCESSING SYSTEM.  AUTHOR: J.FRANK            *
C *  COPYRIGHT (C)1981,1987, WADSWORTH CENTER FOR LABORATORIES AND         *
C *  RESEARCH, NEW YORK STATE DEPARTMENT OF HEALTH, ALBANY, NY 12201.      *
C *  THE CONTENTS OF THIS DOCUMENT ARE PROPRIETARY TO THE CENTER FOR       *
C *  LABORATORIES AND RESEARCH AND ARE NOT TO BE DISCLOSED TO OTHERS OR    *
C *  USED FOR PURPOSES OTHER THAN INTENDED WITHOUT WRITTEN APPROVAL OF     *
C *  THE CENTER FOR LABORATORIES AND RESEARCH                              *
C **************************************************************************
C
C
C
C
C	SUBROUTINE KMEANS.
C
C  THIS ROUTINE WILL PARTITION A NUMBER OF OBJECTS NOBJ, INTO A NUMBER OF
C  CLASSES NCLASS. AN OBJECT "I" BELONGS TO A CLASS "J" IF THE "EUCLIDIAN
C  DISTANCE" OF THIS OBJECT "I" TO THE CENTROID OF THE CLASS "J" IS THE 
C  SMALLEST OF DISTANCES TO ALL CLASS CENTROIDS.
C  A CLASS WILL HAVE MORE OBJECTS IN IT AS ALL OBJECT ARE PROCESSED.
C  THUS A CENTROID COORDINATES ARE CONTINUALLY COMPUTED AND CHANGING.
C  THIS CAUSES AN OBJECT "I" IN ONE CLASS TO FIND ITSELF CLOSER TO 
C  A CENTROID OF ANOTHER CLASS.
C  THAT OBJECT "I" IS REMOVED FROM ONE CLASS AND PUT INTO ANOTHER. 
C
C	REFERENCE: CLUSTER ANALYSIS ALGORITHMS
C		   for data reduction and classification of objects.
C
C		   BY HELMUTH SPATH.
C
C       PUBLISHER: JOHN WILEY & SONS
C                  ELLIS HORWOOD LIMITED,  1980
C
C
C  PAGE 72:
C
C
C
C	THE INITIAL ASSIGNMENT OF THE VECTORS X(NOBJ,NFAC) TO NCLASS CLUSTERS
C	IS GIVEN  BY THE ARRAY P(), WHERE P(I) IS THE CLUSTER NUMBER
C  	OF THE I-th VECTOR; THUS EACH P(I) MUST BE SUCH THAT 
C	1 <= P(I) =< NCLASS AND FOR EACH J=1,...,N AT LEAST ONE I WITH
C	P(I) = J MUST EXIST.
C
C	LET D DENOTE THE SUM OF THE E(J), WHERE E(J) IS THE SUM
C	OF THE SQUARES OF THE DISTANCES BETWEEN THE MEMBERS OF 
C	THE J-th CLUSTER AND THEIR CENTROIDS.
C
C	THE SUBROUTINE MINIMISES D AS FAR AS POSSIBLE BY
C	REPEATED EXCHANGES OF CLUSTER MEMBERS, THE P(I)
C	ARE CORRESPONDINGLY MODIFIED WITHOUT, HOWEVER,
C	CHANGING THE NUMBER OF CLUSTERS.
C
C	THE SUBROUTINE RETURNS THE VALUES AT THE FINAL
C	CONFIGURATION FOR THE CENTROIDS S(NCLASS,NFAC), THE 
C	SUMS E(NCLASS), AND D.
C
C	IF IDR = 1, THE CURRENT VALUES OF D AND OF THE VECTOR P
C	ARE PRINTED AT EACH ITERATION.
C
C	NOBJ	       : NUMBER OF OBJECTS
C       NCLASS         : NUMBER OF CLUSTERS
C       NFAC	       : NUMBER OF DIMENSION IN THE COORD. SYSTEM
C	X(NFAC,NOBJ)   : COORDINATES OF EACH OBJECT
C	P(NOBJ)	       : CLUSTER TO WHICH OBJECT I BELONGS.
C	S(NFAC,NCLASS) : COORDINATES OF THE CENTROIDS OF EACH CLUSTER.
C	E(NCLASS)      : SUM OF THE SQUARED DISTANCES OF ALL OBJECTS 
C		         (IN A CLUSTER) TO THEIR CENTROID.
C	D	       : SUM OF E(NCLASS)
C                     (this has to be reduced after each new clustering)
C	Q(NCLASS)      : NUMBER OF OBJECTS IN EACH CLUSTER.
C
C
C	NOTE:
C		P() HAS BEEN INITIALIZED BEFORE THE CALL TO THIS 
C		ROUTINE.
C
C		IF (NOBJ/NCLASS < 5) SOME GOOD OPTIMAL PARTITIONS ARE
C               NOT FOUND. 
C
C
C
C	NOTE: TO ADD THE POSSIBILITY OF WORKING IN MEMORY OR ON DISK,
C	      I ADDED THE READ OF X() DATA HERE.
C--***************************************************************************

        SUBROUTINE NEWKMEANS(NOBJ, NFAC, X, P, NCLASS, S, E, Q, D, IDR,
     $                       INUM,NIMI,WT, COO, NFTOT, X_DIM, USE_DISK)
        INCLUDE 'CMBLOCK.INC'
        INTEGER P, Q, R, U, V, W, IDR, NFTOT, PT, X_DIM
C	NOTE: IF (USE_DISK) THEN X_DIM=1 ELSE X_DIM=NOBJ
        DIMENSION X(NFAC,X_DIM), P(NOBJ), S(NFAC,NCLASS),INUM(NFAC),
     $      E(NCLASS), Q(NCLASS), COO(NFTOT),NIMI(NOBJ),WT(NFAC)
        LOGICAL DONE,USE_DISK
CNO_SAVE

        IF (NOBJ/NCLASS .LT. 5) THEN
          WRITE(NOUT,2) NOBJ,NCLASS
2         FORMAT(10X,' CL KM COMMAND: NOBJ',I6,' NCLASS ',I6)
          WRITE(NOUT,*) 'NUMBER OF OBJECTS / NUMBER OF CLASSES < 5'
          WRITE(NOUT,*) 'SOME GOOD OPTIMAL PARTITIONS WILL NOT BE FOUND'
        END IF
C
C	INITIALIZE TO ZERO, THE DISTANCES TO THE CENTROIDS AND
C	THE COORDINATES OF THE CENTROIDS.
C
c$doacross if(nfac.ge.64), local(j,k)
        DO J = 1, NCLASS
          Q(J) = 0
          E(J) = 0.
          DO K = 1, NFAC
            S(K,J) = 0.
          END DO
        END DO
C
C	FOR EACH OBJECT, FIND ITS CLUSTER 'R', THE NUMBER OF OBJECT IN THE
C       CLUSTER 'Q(R)' AND ADD ITS COORDINATES 'X(I,K)' TO THE COORDINATES
C	OF THE CENTROID OF THE CLUSTER 'S(R,K)'.
C
        DO I = 1, NOBJ
	  IF(USE_DISK) THEN
            PT = 1
          ELSE
            PT = I
          END IF
C
C	  IN EITHER CASE, WE NEED TO READ X() AT LEAST ONCE.
C
          READ(111)   (COO(J), J = 1, NFTOT),D1,D2,NIMI(I)
c$doacross if(nfac.ge.64), local(j)
          DO J = 1, NFAC
            X(J,PT) = WT(J) * COO(INUM(J))
          END DO

          R = P(I)
          IF (R .LT. 1 .OR. R .GT. NCLASS) THEN
            WRITE(NOUT,*) 'CLUSTER = ',R,' DOES NOT BELONG TO GROUP'
            WRITE(NOUT,*) 
     $      'OBJECT I=',I,' HAS NOT BE PRESET TO ANY CLUSTER'
            RETURN
          END IF
          Q(R) = Q(R) + 1
c$doacross if(nfac.ge.64), local(k)
          DO K = 1, NFAC
            S(K,R) = S(K,R) + X(K,PT)
          END DO
        END DO 

C	REWIND THE UNIT TO READ THE DATA AGAIN       
        IF (USE_DISK) CALL REW(111,1)

C
C	COMPUTE THE CORRECT COORDINATES OF EACH CENTROID BY
C	DIVIDING THE SUM OF COORDS OF ALL ITS OBJECTS BY THE
C	NUMBER OF OBJECTS IN IT.
C
        DO J = 1, NCLASS
          IR = Q(J)
C
C	  EMPTY CLUSTER ? RETURN
C
          IF (IR .EQ. 0) RETURN
          F = 1. / FLOAT(IR)
c$doacross if(nfac.ge.64), local(k)
          DO K = 1, NFAC
            S(K,J) = S(K,J) * F
          END DO
        END DO
C
C	FIND E(R), THE SUM OF DISTANCES OF EACH OBJECT TO THE 
C	CENTROID OF THE CLUSTER 'R' IT BELONGS TO. 
C
        DO I = 1, NOBJ
          R = P(I)
          F = 0.
C
C         COMPUTE THE DISTANCE OF OBJECT 'I' TO THE CENTROID OF 
C         ITS CLUSTER 'R' AND ADD IT TO THE TOTAL SUM OF DISTANCES
C	  'E(R)' IN THAT CLUSTER.
C

          IF(USE_DISK) THEN
            READ(111)   (COO(J), J = 1, NFTOT)
            DO J = 1, NFAC
              X(J,1) = WT(J) * COO(INUM(J))
            END DO
            PT = 1
          ELSE
            PT = I
          END IF

c$doacross if(nfac.ge.64), local(k),reduction(f)
          DO K = 1, NFAC
            F = F + (S(K,R) - X(K,PT))**2
          END DO
          E(R) = E(R) + F
        END DO

C	REWIND THE UNIT TO READ THE DATA AGAIN       
        IF (USE_DISK) CALL REW(111,1)
C
C	COMPUTE THE SUM OF ALL THESE DISTANCES. THE NEW CLUSTERING
C	HAS TO REDUCE THIS D VARIABLE OR STOP AND RETURN.
C
        D = 0.
        DO J = 1, NCLASS
          D = D + E(J)
        END DO
C
C	FOR EACH OBJECT 'I', CHECK ITS DISTANCE TO THE CENTROID OF
C	THE OTHER CLUSTERS. IF THIS DISTANCE IS SMALLER THAN THE DISTANCE
C	OF THIS OBJECT 'I' TO ITS OWN CLUSTER'S CENTROID, THEN MOVE
C	THE OBJECT 'I' TO THE NEW CLUSTER. STOP WHEN NO MORE OBJECTS 
C	ARE MOVED.
C
        I = 0
        IT = 0

C	LOOP UNTIL DONE AND ALL OBJECTS HAVE BEEN CHECKED (I.GT.NOBJ) 
	DONE = .TRUE.

10      I = I + 1
        IF (I .GT. NOBJ) THEN
          IF (DONE) RETURN
          DONE = .TRUE.

C         REWIND THE UNIT TO READ THE DATA AGAIN       
          IF (USE_DISK) CALL REW(111,1)
          I = 1
        END IF

        IF(USE_DISK) THEN
          READ(111)   (COO(J), J = 1, NFTOT)
c$doacross if(nfac.ge.64), local(j)
          DO J = 1, NFAC
            X(J,1) = WT(J) * COO(INUM(J))
          END DO
          PT = 1
        ELSE
          PT = I
        END IF
C
C	FIND THE CLUSTER, R, AND THE NUMBER, U, OF OBJECTS IN IT.
C       NOTE THAT IF THERE IS ONLY ONE OBJECT, IT IS CLOSEST TO ITSELF.
C
        R = P(I)
        U = Q(R)
        IF (U .LE. 1) GOTO 10
C
C	COMPUTE DISTANCE 'A' OF OBJECT 'I' TO ITS OWN CLUSTER 'R'
C
        H = FLOAT(U)
        H = H / (H - 1.)
        F = 0.
c$doacross if(nfac.ge.64), local(k),reduction(f)
        DO K = 1, NFAC
          F = F + (S(K,R) - X(K,PT))**2
        END DO
        A = H * F
        B = 1. E30
C
C	FIND DISTANCE 'F' OF THE SAME OBJECT 'I' TO THE OTHER CLUSTERS.
C	DISTANCE 'B' IS THE SMALLEST OF THESE DISTANCES 'F' CORRESPONDING
C	TO A CLUSTER 'V' WHICH HAS 'W' NUMBER OF OBJECTS IN IT.
C
        DO J = 1, NCLASS
          IF (J .NE. R) THEN
            U = Q(J)
            H = FLOAT(U)
            H = H / (H + 1.)
            F = 0.
c$doacross if(nfac.ge.64), local(k),reduction(f)
            DO K = 1, NFAC
              F = F + (S(K,J) - X(K,PT))**2
            END DO
            F = H * F
            IF (F .LE. B) THEN
C
C             SMALLER DISTANCE FOUND, SAVE THAT CLUSTER 'V', THE NUMBER 
C	      OF OBJECTS IN IT AND ITS DISTANCE TO 'I'.
C
              B = F
              V = J
              W = U
            END IF
          END IF          
        END DO
 
        IF (B .GE. A) THEN
C
C	  OBJECT 'I' IS CLOSEST TO CLUSTER 'R', TAKE NEXT OBJECT.
C
          IT = IT + I
          GOTO 10
        END IF

C
C	OBJECT 'I' IS CLOSEST TO CLUSTER 'V' INSTEAD. SO TAKE OBJECT
C	'I' FROM CLUSTER 'R' AND PUT IT IN CLUSTER 'V'
C
	DONE = .FALSE.
C
C       RECOMPUTE THEIR CENTROID AND SUM OF DISTANCES.
C       these are equations (3.2.16, 3.2.17, 3.2.20 and 3.2.21) on page
C	71-72 of the above book reference.
C
        IT = 0
        E(R) = E(R) - A
        E(V) = E(V) + B
        D = D - A + B
        H = FLOAT(Q(R))
        G = FLOAT(W)
        A = 1. / (H - 1.)
        B = 1. / (G + 1.)
C
C	  RECOMPUTE COORDS. OF CENTROID OF 'R' AND 'V' CLUSTERS
C
c$doacross if(nfac.ge.64), local(k,f)
        DO K = 1, NFAC
          F = X(K,PT)
C
C         OBJECT 'I' REMOVED FROM CLUSTER 'R'.
C
          S(K,R) = ( H * S(K,R) - F) * A
C
C	  OBJECT 'I' ADDED TO CLUSTER 'V'
C
          S(K,V) = ( G * S(K,V) + F) * B
        END DO
C
C	RESET THE CLUSTER NUMBER FOR OBJECT 'I', RECOMPUTE THE
C	NUMBER OF ELEMENTS IN CLUSTER 'R' AND 'V'.
C
        P(I) = V
        Q(R) = Q(R) - 1
        Q(V) = Q(V) + 1
C
C       PRINT AN OUTPUT
C
        IF( IDR .EQ. 1) WRITE(6,16) I, D
        IF( IDR .EQ. 1) WRITE(6,17) (U,P(U), U = 1, NOBJ)
16      FORMAT(2X,I4,' MIMINIZE DISTANCE D = ',1X,F12.2,/,
     $         'CLUSTER MEMBERSHIP',/)
17      FORMAT(10X,20I3,/)

        GOTO 10
   
        END
