/*$Header: /net/ithaca/usr1/web/src/RCS/getdoc.c,v 1.6 1996/07/01 18:14:04 mladjadj Exp $*/
/*
*****************************************************************************
**    getdoc.c
**
** **************************************************************************
** *  SPIDER - MODULAR IMAGE PROCESSING SYSTEM.  AUTHOR: J.FRANK            *
** *  COPYRIGHT (C)1981,1993, WADSWORTH CENTER FOR LABORATORIES AND         *
** *  RESEARCH, NEW YORK STATE DEPARTMENT OF HEALTH, ALBANY, NY 12201.      *
** *    THE CONTENTS OF THIS DOCUMENT ARE PROPRIETARY TO THE CENTER FOR     *
** *    LABORATORIES AND RESEARCH AND ARE NOT TO BE DISCLOSED TO OTHERS OR  *
** *    USED FOR PURPOSES OTHER THAN INTENDED WITHOUT WRITTEN APPROVAL OF   *
** *    THE CENTER FOR LABORATORIES AND RESEARCH 			    *
** **************************************************************************
**
**    PURPOSE:  RECOVERS DOCUMENT FILE INFO.
**
**    PARAMETERS:
**         FPDOC         DOC. FILE FILE POINTER               (SENT)
**         DOCNAM        DOC. FILE NAME                       (SENT)
**         DOCEXT        DOC. FILE EXTENSION                  (SENT)
**         MAXKEY        NUMBER OF ROWS IN DBUF               (SENT)
**         MAXREG        NUMBER OF COLUMNS IN DBUF            (SENT)
**         DBUF          BUFFER OF RETRIEVAL FROM DOC. FILE   (RETURNED)
**         LASTKEY       NUMBER OF HIGHEST KEY FOUND IN FILE  (RETURNED)
**
**    RETURNS:
**
**         LERR          ERROR FLAG (  0 IS NORMAL)           
**
**                       -5   EXCESSIVE REGISTERS FOR >= ONE KEY
**                       -4   KEY > MAXKEY FOR >= ONE KEY
**                        1   LACK SUFFICIENT REGISTER FOR >= ONE KEY
**                       20   MEMORY ALLOCATION FAILED
**                       21   NEW DOC FILE NAME TOO LONG
**                       10   NEW DOC FILE CAN NOT BE OPENED
**                       22   ERROR READING DOC FILE LINE
**                       30   MAXREGP1 TOO LARGE
**
**    TYPICAL DOC FILE LINE:
**    COL: 123456789 123456789 123456789 123456789 123456789 1234565789
**           1 4   20.070000   17.000000   17.000000   17.000000
**         KEY #REGS/LINE    VALUES ........
**
**    WHEN RETRIEVED THIS LINE IS PLACED IN DBUF STARTING AT 
**    [(KEY -1)*MAXREG].  THE VALUES PLACED THERE ARE THE NUMBER OF
**    REGISTERS RETIVED FOR THAT KEY, FOLLOWED BY VALUES OF THE
**    REGISTERS.  THUS FOR THE ABOVE SAMPLE AT DBUF[0] WE HAVE
**    2, 20, 17, 17, 17.
**
**    TO USE DBUF, COMPUTE ADDRESS OF EACH ENTRY BY USING POINTERS:
**
**    FOR J REG IN  KEY3 :    PT = DBUF + (KEY3 - 1) * MAXREG + J;
**
*****************************************************************************
*/

#include "std.h"
#include "routines.h"

/***************************  getdoc    *******************************/

 int getdoc(FILE * fpdoc, char *docnam, char * docext,  
            int maxkeys, int maxregp1, float **buf, int * lastkeygot)
 {
  char        outstr[109], docname[81], reclin[101];
  FILE      * ptrdoc;
  float     * dbuf, *lptr;
  int         regsonline, items, message1,  message2, lerr, keygot, i;
  float       regis[6];
  int         message0;

  if (maxregp1 > 7)
     {spout("*** Pgm. error in getdoc usage"); return 30; }  

  *lastkeygot = 0;
  message0    = TRUE;
  message1    = TRUE;
  message2    = TRUE;
  lerr        = 0;     

  /*  Add datexc (name extension ) to docnam if not already there  */
  if (strlen(docnam) > (size_t) 81)
     {spouts("*** Doc. file name too long: "); spout(docnam); return 21;}
  
  strcpy(docname,docnam);

  if (docext)
     {
     if ((strlen(docname) + strlen(docext) + 1) > (size_t) 80)
        {spouts("*** Doc. file name too long: "); spout(docname); 
        return 21; }  
     strcat(strcat(docname,"."),docext);
     }

  /* open a new file if fpdoc is NULL, close it upon return */
  ptrdoc = fpdoc;
  if (!ptrdoc && ((ptrdoc = fopen(docname,"r")) == NULL))
     {spouts("*** Unable to open docfile: "); spout(docname); return 10;}
  else
     { /* already have open doc file */
     rewind(ptrdoc);
     }
 
  /* print document file name */
  spouts("Reading docfile: "); spout(docname);

  /* run time allocation of memory space for buf */
  if ((dbuf = (float *) calloc(maxkeys * maxregp1, sizeof(float)))
           == (float *) NULL)
     {spout("*** Memory allocation failed in getdoc!"); 
      fclose(ptrdoc); return 20; }


  /* read next line (until end-of-line or 80 char) from doc file */
  while ((fgets(reclin,100,ptrdoc)) != (char *) NULL)
     {
     if (ferror(ptrdoc))
        { /*  fatal error */
        spouts("*** Error reading docfile line: "); spout(reclin);
        fclose(ptrdoc); return 22;
        }

     if ((strstr(reclin,";")) == (char *) NULL)
        { /* skip any  comment line */ 
        items = sscanf(reclin,"%d %1d%12f%12f%12f%12f%12f%12f",
                &keygot, &regsonline, &regis[0],&regis[1],&regis[2],
                                    &regis[3],&regis[4],&regis[5]);
        /**************
        printf(" items: %d, key: %5d regs: %d %f %f %f \n",
                 items,keygot,regsonline, 
                 regis[0],regis[1],regis[2]);
        ****************/

        if(items > 0) 
	{
	if (items < 3 || regsonline < maxregp1 - 1 && message0) 
           {
           /* not enough registers on this line, possible error */
           sprintf(outstr, 
              "*** Lacks registers for one or more key, e.g. key: %d",
                        keygot);
           spout(outstr); lerr = 1; message0 = FALSE;       
           } 

        if (keygot > maxkeys) 
           {
           if (message1)
              {
              /* this key can not be stored in array, possible error */
              sprintf(outstr,"*** Doc file key: %d > max keys: %d",
                      keygot,maxkeys);
              spout(outstr); lerr = -4; message1 = FALSE; 
              }
           keygot = 0;      
           } 

        if (regsonline > (maxregp1 - 1))  
           {  /* doc file has more registers than are retrieved */
           if (message2)
              {
              sprintf(outstr, 
              "*** Unused registers on one or more key, e.g. key: %d",
                        keygot);
              spout(outstr);  lerr = -5; message2 = FALSE;
              }
           /* even though regsonline was >= maxregp1, continue */
           regsonline = maxregp1 - 1;
           }
   
        if (keygot > 0)
           {  /* put key contents in dbuf */ 

           /* find pointer into dbuf */
           lptr = dbuf + (keygot - 1) * maxregp1;

           /* store number of registers for key in dbuf */
           *(lptr++) = regsonline;

           for (i = 0; i < regsonline; i++)
              {
              /* read the register entries */
              *(lptr++) = regis[i]; 
              }
               
           *lastkeygot = MYMAX(*lastkeygot, keygot);
           } /* (keygot > 0)                        */
	   } /* eof, blank line etc...      ML      */
        }    /* end of      if ((strstr(reclin..... */
     }       /* while (read next line of doc file)  */

  /* close the doc file  */
  if (!fpdoc) 
     {fclose(ptrdoc);}
  else
     { /* poition file at end for append */
     fseek(fpdoc,0L,SEEK_END);
     }

  /* the doc file has been successfully read */
  *buf  = dbuf;

  return lerr;
  }

